use std::{boxed::Box, io::{Read, Write}, path::PathBuf};

pub type Output    = Box<dyn Write>;
pub type Input     = Box<dyn Read>;
pub type Result<T> = std::result::Result<T, Error>;

pub struct Error {
    pub exit_code:  i32,
    pub text:       String,
}

impl Error {
    pub fn new(exit_code: i32, text: String) -> Error {
        Error{exit_code, text}
    }

    pub fn try_with_code<T, S>(result: std::result::Result<T, S>, exit_code: Option<i32>) -> Result<T> where S: std::fmt::Display {
        match result {
            Ok(value)  => Ok(value),
            Err(error) => Err(Error::new(Self::get_exit_code(exit_code), error.to_string())),
        }
    }

    pub fn try_with<T, S>(result: std::result::Result<T, S>) -> Result<T> where S: std::fmt::Display {
        Self::try_with_code(result, None)
    }

    fn get_exit_code(exit_code: Option<i32>) -> i32 {
        match exit_code {
            Some(exit_code) => exit_code,
            None            => -1,
        }        
    }
}

pub fn open_output(output_file: Option<PathBuf>) -> Result<Output> {
    match output_file {
        Some(output_path) => Ok(Box::new(Error::try_with(std::fs::File::create(output_path))?)),
        None              => Ok(Box::new(std::io::stdout())),
    }
}

pub fn open_input(input_file: Option<PathBuf>) -> Result<Input> {
    match input_file {
        Some(input_path) => Ok(Box::new(Error::try_with(std::fs::File::open(input_path))?)),
        None             => Ok(Box::new(std::io::stdin())),
    }
}